/*******************************************************************************
* Copyright (C) 2014 - 2015 Intel Corp. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Intel Corp. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Intel Corp. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

using System;
using System.Linq;
using System.ServiceProcess;
using System.IO;
using System.Text;

namespace FwUpdateAPI
{
    /// <summary>
    /// This class includes various useful utility functions, some for internal use and some for
    /// applications to use
    /// </summary>
    public class Utilities
    {
        /// <summary>
        /// Reads the NVM version from FW image file
        /// </summary>
        /// <param name="path">Path to FW image file to read from</param>
        /// <returns>NVM version as found in the file</returns>
        /// <remarks>
        /// For printing the version, refer to GetCurrentNvmVersionString() below or
        /// Utilities.NvmVersionToString()
        /// </remarks>
        public static UInt32 GetImageNvmVersion(string path)
        {
            const uint nvmVersionOffset = 0xA;

            using (var fs = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                var size = (int)fs.Length;
                var reader = new BinaryReader(fs);

                var imageBuffer = reader.ReadBytes(size);
                var nvmStartOffset = new DeviceFwInfo(new FileFwInfoSource(imageBuffer)).GetSectionInfo()[Sections.Digital].Offset;
                return imageBuffer[nvmVersionOffset + nvmStartOffset];
            }
        }

        /// <summary>
        /// Converts NVM version to string representation, assuming the version is valid
        /// </summary>
        /// <param name="version">NVM version to convert to string</param>
        /// <remarks>
        /// NVM number should be parsed like it's a hex number, but displayed like a decimal
        /// number, e.g. for rev 10 you get the value 16 (0x10) from the FW.
        /// </remarks>
        /// <returns>Strings representation of the NVM version</returns>
        public static string NvmVersionToString(UInt32 version)
        {
            return version.ToString("X");
        }

        /// <summary>
        /// Allows calling get current version functions in a safe manner which
        /// returns "N/A" when in safe mode
        /// </summary>
        /// <param name="func">Function object that reads current version and returns it as a string</param>
        /// <returns>Version string or "N/A" if the controller is safe-mode</returns>
        /// <exception>
        /// Rethrows any error thrown from func that isn't about safe-mode
        /// </exception>
        public static string SafeGetVersion(Func<string> func)
        {
            try
            {
                return func();
            }
            catch (TbtException e)
            {
                switch (e.ErrorCode)
                {
                    case TbtStatus.INVALID_OPERATION_IN_SAFE_MODE:
                    case TbtStatus.SDK_INVALID_OPERATION_IN_SAFE_MODE:
                        return "N/A";
                    default:
                        throw;
                }
            }
		}

        /// <summary>
        /// Reads the PD version from FW image file
        /// </summary>
        /// <param name="path">Path to FW image file to read from</param>
        /// <returns>PD version as found in the file</returns>
        /// <remarks>
        /// Function returns "N/A" if no PD version is found. 
        /// The "permanent string" ("TPS65982 HW     FW") can only be found 
        /// in image files that contain available PD versions.
        /// This example currently works only with the Texas Instrument (TI) 
        /// chip TPS65982. This should be implemented according to each specific 
        /// PD solution.
        /// </remarks>
        public static string GetImagePdVersion(string path)
        {
            using (var fs = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                const int notFound = -1;
                const int versionFromFileLength = 10;
                const string permanentValue = "TPS65982 HW     FW";
                var version = "N/A";
                var reader = new BinaryReader(fs);
                var imageString = Encoding.ASCII.GetString(reader.ReadBytes((int) fs.Length));

                var offset = imageString.IndexOf(permanentValue);
                if (offset == notFound)
                    return version;

                offset += permanentValue.Length;
                version = imageString.Substring(offset, versionFromFileLength);

                // Removing leading 0's in major version so that instead of 
                // (for example) version being 0001.01.00, version will be 1.01.00
                // (if major version is 0, version will be (for example) 0.08.12).
                // (Side note: the reason we don't use the "Version" class here is 
                // because we may recieve hex values as part of the version.)
                version = version.TrimStart('0');

                if (version[0] == '.')
                {
                    version = "0" + version;
                }

                return version;
            }
        }


        /// <summary>
        /// Checks if a host controller supports device FW update
        /// </summary>
        /// <param name="controllerId">Controller ID as comes from WMI</param>
        /// <returns>True if the controller supports device FW update; otherwise - false</returns>
        public static bool SupportsDeviceUpdate(string controllerId)
        {
            return GetHwConfiguration(controllerId).Generation >= HwGeneration.DSL6540_6340;
        }

        /// <summary>
        /// Checks if a host controller is supported for FW update
        /// </summary>
        /// <param name="controllerId">Controller ID as comes from WMI</param>
        /// <returns>True if the controller is supported for FW update; otherwise - false</returns>
        public static bool HostUpdateSupported(string controllerId)
        {
            return GetHwConfiguration(controllerId).Generation > HwGeneration.DSL4510_4410;
        }

        /// <summary>
        /// Get HW configuration information from controller ID; helper function
        /// </summary>
        /// <param name="controllerId">Controller ID as comes from WMI</param>
        /// <remarks>
        /// Controller ID format is something like:
        /// <![CDATA[
        /// PCI\VEN_8086&DEV_156A&SUBSYS_00000000&REV_00\6&1e803622&0&000000E4_0
        /// ]]>
        /// The 4 hex-digits after "DEV_" are the device ID
        /// </remarks>
        /// <returns>HWInfo object with HW configuration</returns>
        private static HwInfo GetHwConfiguration(string controllerId)
        {
            const string idStartString = "DEV_";
            const int idLength = 4;

            var idLocation = controllerId.IndexOf(idStartString) + idStartString.Length;
            var idString = controllerId.Substring(idLocation, idLength);
            var id = UInt16.Parse(idString, System.Globalization.NumberStyles.AllowHexSpecifier);
            return FwInfoSource.HwConfiguration(id);
        }

        /// <summary>
        /// Check if a pointer comes from FW is valid
        /// </summary>
        /// <param name="pointer">Pointer value as comes from FW</param>
        /// <param name="pointerSize">
        /// The real size of the pointer (to be able to handle 24-bit pointer)
        /// </param>
        /// <returns>True if the pointer is valid, otherwise - false</returns>
        /// <remarks>Invalid pointer is NULL pointer or -1 (0xFFFFFFFF in uint)</remarks>
        internal static bool ValidPointer(uint pointer, int pointerSize)
        {
            var invalidPointer = BitConverter.GetBytes(pointer).Take(pointerSize).All(b => b == byte.MaxValue);
            return pointer != 0 && !invalidPointer;
        }

        private const string ServiceName = "ThunderboltService";

        /// <summary>
        /// Tries to start the service if it exists
        /// </summary>
        /// <returns>
        /// True if the service is up and running; false if the service doesn't exist or we don't
        /// know if it succeeded to start
        /// </returns>
        internal static bool StartService()
        {
            var serviceController = GetServiceController();

            // If the service doesn't exist at all, no need to start it
            if (serviceController == null)
                return false;

            // If service is up and running just exit gracefully
            if (serviceController.Status == ServiceControllerStatus.Running)
                return true;

            // If the service exists but is not started, go ahead
            try
            {
                serviceController.Start();
                // Wait for a second for the service to come up
                serviceController.WaitForStatus(ServiceControllerStatus.Running, new TimeSpan(TimeSpan.TicksPerSecond));
                return true;
            }
            catch
            {
                return false;
            }
        }

        /// <summary>
        /// Helper function for StartSerivce()
        /// </summary>
        private static ServiceController GetServiceController()
        {
            if (IsServiceExists())
                return new ServiceController(ServiceName);
            throw new TbtException(TbtStatus.SDK_SERVICE_NOT_FOUND);
        }

        /// <summary>
        /// Helper function for StartSerivce()
        /// </summary>
        private static bool IsServiceExists()
        {
            var list = ServiceController.GetServices();
            var tbtServiceController = list.FirstOrDefault(s => s.ServiceName == ServiceName);
            return (tbtServiceController != null);
        }

    }
}
