﻿/*******************************************************************************
* Copyright (C) 2014-2015 Intel Corp. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Intel Corp. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Intel Corp. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

using System;
using System.ComponentModel;
using System.Windows;
using System.Windows.Media.Animation;
using System.Timers;

namespace FwUpdateTool.WizardSteps
{
    /// <summary>
    /// Interaction logic for FwUpdateProcessScreen.xaml
    /// In this form a progress bar is implemented. The progress bar does not reflect the actual progress of flashing it just fills after two minutes. 
    /// ProgBar.maximum property indicates time (in units of 0.1 of a second) to fill the bar. E.g. ProgBar.maximum = 1200 means that the bar will be filled after 2 minutes.
    /// </summary>
    public partial class FwUpdateProcessScreen : IWizardScreen, INotifyPropertyChanged
    {
        readonly DoubleAnimation _ani = new DoubleAnimation(1, 0.3, TimeSpan.FromMilliseconds(1000));
        // A timer to increment the progress bar. Uses 0.1 of a second units
        private readonly Timer _timer = new Timer(100);

        public FwUpdateProcessScreen()
        {
            InitializeComponent();
            InitializeFlashAnimation();
            _timer.Elapsed += OnTimedEvent;
        }

        /// <summary>
        /// Initialize the animation that appears during the FW update process
        /// </summary>
        private void InitializeFlashAnimation()
        {
            _ani.AutoReverse = true;
            _ani.Completed += ani_Completed;
            Logo.BeginAnimation(OpacityProperty, _ani);
        }

        /// <summary>
        /// restart the animation once completed 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void ani_Completed(object sender, EventArgs e)
        {
            Logo.BeginAnimation(OpacityProperty, _ani);
        }

        /// <summary>
        /// Handle progress bar increment.
        /// In case flashing takes more than the defined Maximum the bar will return to its half filled state.
        /// This function is called with the elapsed event of the timer.
        /// </summary>
        private void OnTimedEvent(Object source, ElapsedEventArgs e)
        {
            Dispatcher.Invoke(
            () =>
            {
                if (ProgBar.Value < ProgBar.Maximum)
                    ProgBar.Value++;
                else
                    ProgBar.Value = ProgBar.Maximum / 2;
            });
        }

        #region IWizardScreen
        public bool CancelButtonActive
        {
            get { return false; }
        }

        public bool NextButtonActive
        {
            get { return false; }
        }

        public bool BackButtonActive
        {
            get { return false; }
        }

        public string[] Title
        {
            get { return new[] {"Flashing..."}; }
        }
        #endregion

        #region INotifyPropertyChanged
        public event PropertyChangedEventHandler PropertyChanged;
        // Create the OnPropertyChanged method to raise the event 
        public void OnPropertyChanged(string name)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null)
            {
                handler(this, new PropertyChangedEventArgs(name));
            }
        }

        protected void OnPageChange()
        {
            OnPropertyChanged("CancelButtonActive");
            OnPropertyChanged("NextButtonActive");
            OnPropertyChanged("BackButtonActive");
        }
        #endregion

        /// <summary>
        /// Start progress bar fill once this form is reached.
        /// </summary>
        private void UserControl_IsVisibleChanged(object sender, DependencyPropertyChangedEventArgs e)
        {
            _timer.Start();
        }
    }
}
